import numexpr as ne
import sympy
from sympy import *
import numpy as np
import matplotlib.pyplot as plt
import re
# from matplotlib import cm
# from mpl_toolkits.axes_grid1 import make_axes_locatable


class SearchExtremes:
    def find(x: str, y: str, z: str, x_from: float, x_to: float, y_from: float, y_to: float):
        """
        Функция находит экстермумы, седловые точки и точки дальнейшего исследования и строит 3Д график

        Parameters
        ===========
        :param x: str
            символ х
        :param y: str
            символ у
        :param z: str
            строковая функция
        :param x_from: float
            интервал "от" по х
        :param x_to: float
            интервал "до" по х
        :param y_from: float
            интервал "от" по у
        :param y_to: float
            интервал "до" по у

        Returns
        ===========
        3Д график с отмеченными точками экстремумов
        """

        # Преобразуем произвольное выражение в тип, который можно использовать внутри SymPy
        func = sympify(z)
        # Считаем первые производные по х и у и приравниваем их к нулю чтобы найти эти х и у
        values_diff = solve([func.diff(x), func.diff(y)], dict=True)  # делаем это все словарем (dict=True)
        v_d = []

        # проверяем есть ли пара в словаре для икса и игрика и при ее отсутствии добавляем
        for i in range(len(values_diff)):
            if (len(values_diff[i]) == 1) and (value == 0 for value in values_diff[i].values()):
                values_diff[i].update(dict(n=0))


        # заводим цикл по длине values_diff, чтобы избавиться от значений словаря
        for i in range(len(values_diff)):
            v_d_in = []  # список для каждой отдельной точки
            for key, value in values_diff[i].items():
                v_d_in.append(value)

            if (type(v_d_in[0]) != int) and (type(v_d_in[1]) != int):
                if ((v_d_in[0]).has(I) is False) and ((v_d_in[1]).has(I) is False):  # проверяем на налич комплекс чисел
                    v_d_in[0] = float(v_d_in[0])  # Переводим во float для удобства
                    v_d_in[1] = float(v_d_in[1])
                    v_d.append(v_d_in)
        values_diff = v_d  # список точек, где на первом месте х, на втором у

        # Создаем гессиан
        ges = np.array([[func.diff(x).diff(x), func.diff(x).diff(y)],
                        [func.diff(y).diff(x), func.diff(y).diff(y)]])

        points_max = []  # список максимумов
        points_min = []  # список минимумов
        points_not_defined = []  # список точек требующих доп исследования
        saddle_point = []  # список седловых точек

        for i in range(len(values_diff)):  # создаем цикл перебирающий каждую точку

            # проверяем значения гессиана (a, b, c) на то что они являются числами и если да, то считаем их,
            # а если нет то идем дальше
            if (type(ges[0][0]) == sympy.core.numbers.Integer) or (type(ges[0][0]) == sympy.core.numbers.Zero):
                a = ges[0][0]
            else:
                a = func.diff(x).diff(x).subs([(x, values_diff[i][0]), (y, values_diff[i][1])])
            if (type(ges[0][1]) == sympy.core.numbers.Integer) or (type(ges[0][1]) == sympy.core.numbers.Zero):
                b = ges[0][1]
            else:
                b = func.diff(x).diff(y).subs([(x, values_diff[i][0]), (y, values_diff[i][1])])
            if (type(ges[1][1]) == sympy.core.numbers.Integer) or (type(ges[1][1]) == sympy.core.numbers.Zero):
                c = ges[1][1]
            else:
                c = func.diff(y).diff(y).subs([(x, values_diff[i][0]), (y, values_diff[i][1])])

            # проверка на существование локальных экстремумов по формуле (a * c - b ** 2) > 0

            if (values_diff[i][0] >= x_from) and (values_diff[i][0] <= x_to) and (values_diff[i][1] >= y_from) and (
                    values_diff[i][1] <= y_to):
                if (a * c - b ** 2) > 0:

                    if a > 0:
                        # Локальный экстремум(минимум)
                        points_min.append((values_diff[i][0], values_diff[i][1]))
                    elif a < 0:
                        # Локальный экстремум(максимум)
                        points_max.append((values_diff[i][0], values_diff[i][1]))
                elif (a * c - b ** 2) == 0:
                    # Требуется дополнительное исследование (a*c-b**2) = 0
                    points_not_defined.append((values_diff[i][0], values_diff[i][1]))

                # Седловая точка
                if ((a > 0) and (a * c < 0)) or ((a < 0) and (a * c < 0)):
                    saddle_point.append((values_diff[i][0], values_diff[i][1]))

        # если нет локального экстремума выводим запись о его отсутствии, если есть выводим точки
        # аналогично с точкой доп исследования и с седловой
        #if not points_max:
        #    print("Локального экстремума (максимума) нет")
        #else:
        #    print("Локальный экстремум (максимум)")
        #    for i in range(len(points_max)):
        #        print(points_max[i])
        #if not points_min:
        #    print("Локального экстремума (минимума) нет")
        #else:
        #    print("Локальный экстремум (минимум)")
        #    for i in range(len(points_min)):
        #        print(points_min[i])
        #if not points_not_defined:
        #    print("Дополнительное исследование не требуется")
        #else:
        #    print("Требуется дополнительное исследование")
        #    for i in range(len(points_not_defined)):
        #        print(points_not_defined[i])
        #if not saddle_point:
        #    print("Седловой точки нет")
        #else:
        #    print("Седловая точка")
        #    for i in range(len(saddle_point)):
        #        print(saddle_point[i])

        # Для 3д графика
        # Построение списков точек минимума для графика

        all_points_x = []
        all_points_y = []

        lst_points_x_min = []
        lst_points_y_min = []
        for i in range(len(points_min)):
            if points_min:
                lst_points_x_min.append(float(points_min[i][0]))
                lst_points_y_min.append(float(points_min[i][1]))
                # Считаем точки по x и y для центрирования по графику
                all_points_x.append(float(points_min[i][0]))
                all_points_y.append(float(points_min[i][1]))

        lst_points_z_min = []

        for i in range(len(lst_points_x_min)):
            lst_points_z_min.append(float(func.subs([(x, lst_points_x_min[i]), (y, lst_points_y_min[i])])))

        # Построение списков точек максимума для графика
        lst_points_x_max = []
        lst_points_y_max = []
        for i in range(len(points_max)):
            if points_max:
                lst_points_x_max.append(float(points_max[i][0]))
                lst_points_y_max.append(float(points_max[i][1]))
                # Считаем точки по x и y для центрирования по графику
                all_points_x.append(float(points_max[i][0]))
                all_points_y.append(float(points_max[i][1]))

        lst_points_z_max = []

        for i in range(len(lst_points_x_max)):
            lst_points_z_max.append(float(func.subs([(x, lst_points_x_max[i]), (y, lst_points_y_max[i])])))

        # Построение списков седловой точки
        lst_points_x_saddle = []
        lst_points_y_saddle = []
        lst_points_z_saddle = []

        for i in range(len(saddle_point)):
            if saddle_point:
                lst_points_x_saddle.append(float(saddle_point[i][0]))
                lst_points_y_saddle.append(float(saddle_point[i][1]))
                # Считаем точки по x и y для центрирования по графику
                all_points_x.append(float(saddle_point[i][0]))
                all_points_y.append(float(saddle_point[i][1]))

        for i in range(len(lst_points_x_saddle)):
            lst_points_z_saddle.append(float(func.subs([(x, lst_points_x_saddle[i]), (y, lst_points_y_saddle[i])])))

        # Построение списков точки доп исследования
        lst_points_x_dop = []
        lst_points_y_dop = []
        lst_points_z_dop = []

        for i in range(len(points_not_defined)):
            if points_not_defined:
                lst_points_x_dop.append(float(points_not_defined[i][0]))
                lst_points_y_dop.append(float(points_not_defined[i][1]))
                # Считаем точки по x и y для центрирования по графику
                all_points_x.append(float(points_not_defined[i][0]))
                all_points_y.append(float(points_not_defined[i][1]))

        for i in range(len(lst_points_x_dop)):
            lst_points_z_dop.append(float(func.subs([(x, lst_points_x_dop[i]), (y, lst_points_y_dop[i])])))

        # Среднее x, y для центрирования
        if points_not_defined or saddle_point or points_max or points_min:
            mean_x = np.mean(all_points_x)
            mean_y = np.mean(all_points_y)
            length_x = mean_x + abs(max(all_points_x) - mean_x) + abs(max(all_points_x) - min(all_points_x))/5 + 1
            length_y = mean_y + abs(max(all_points_y) - mean_y) + abs(max(all_points_y) - min(all_points_y))/5 + 1
            x_from = mean_x - length_x
            x_to = mean_x + length_x
            y_from = mean_y - length_y
            y_to = mean_y + length_y


        # Подпись осей
        fig = plt.figure(figsize=(10, 8))
        ax = fig.add_subplot(projection='3d')
        ax.set_title('Поиск экстремумов')
        ax.set_zlabel('Z')
        ax.set_ylabel('Y')
        ax.set_xlabel('X')

        x = np.linspace(x_from,x_to, 25)
        y = np.linspace(y_from,y_to, 25)
        x, y = np.meshgrid(x, y)
        Z = ne.evaluate(z)

        surf = ax.plot_surface(x, y, Z, cmap='viridis', shade=True, alpha=0.5)
        fig.colorbar(surf, shrink=0.5, aspect=10)

        # Рисуем на графике точки минимума и максимума
        for i in range(len(lst_points_x_min)):
            ax.scatter(lst_points_x_min[i], lst_points_y_min[i], lst_points_z_min[i], c='r', s=50,
                       label=f'Точка минимума {lst_points_x_min[i], lst_points_y_min[i], lst_points_z_min[i]}')

        # Рисуем на графике точки максимума
        for i in range(len(lst_points_x_max)):
            ax.scatter(lst_points_x_max[i], lst_points_y_max[i], lst_points_z_max[i], c='b', s=50,
                       label=f'Точка максимума {lst_points_x_max[i], lst_points_y_max[i], lst_points_z_max[i]}')

        # Рисуем на графике седловые точки
        for i in range(len(lst_points_x_saddle)):
            ax.scatter(lst_points_x_saddle[i], lst_points_y_saddle[i], lst_points_z_saddle[i], c='black', s=50,
                       label=f'Седловая точка {lst_points_x_saddle[i], lst_points_y_saddle[i], lst_points_z_saddle[i]}')

        # Рисуем на графике точки доп исследования
        for i in range(len(lst_points_x_dop)):
            ax.scatter(lst_points_x_dop[i], lst_points_y_dop[i], lst_points_z_dop[i], c='g', s=50,
                       label=f'Точка, требующая доп исследований {lst_points_x_dop[i], lst_points_y_dop[i], lst_points_z_dop[i]}')

        # Рисуем легенду
        if points_max or points_min or saddle_point or points_not_defined:
            if not points_max:
                ax.scatter(0, 0, 0, c='w', s=0.1, label=f'Точки максимума не найдены')
        if not points_min:
            ax.scatter(0, 0, 0, c='w', s=0.1, label=f'Точки минимума не найдены')
        if not saddle_point:
            ax.scatter(0, 0, 0, c='w', s=0.1, label=f'Седловые точки не найдены')
        if not points_not_defined:
            ax.scatter(0, 0, 0, c='w', s=0.1, label=f'Точки для дальнейшего исследования не найдены')
            ax.legend(labelcolor='black')
        else:
            ax.scatter(0, 0, 0, c='w', s=0.1, label=f'Никакие точки не найдены')
            ax.legend(labelcolor='black')

        plt.show()



import numpy as np
import pandas as pd
import sklearn
import time
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn import metrics
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
import numpy as np
import pandas as pd
import sklearn
from sklearn.svm import SVC
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn import metrics
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import classification_report, confusion_matrix


class Сlassification:
    def logistic(x_, y_):

        """
        Функция, реализующая модель классификации на два класса на основе логистической регрессии.
        Parameters
        ----------
        x_, y_: numpy array
            Массивы обучающей переменной и предсказываемой
        Returns
        -----------
        Accuracy, Precision, Recall, Коэффициенты регрессии, Массив предсказанных значений, График классификации
        """

        start = time.time()

        #         Массивы обучающей переменной и предсказываемой, разделенные на обучающую и тестовую выборки
        x_train, x_test, y_train, y_test = train_test_split(x_, y_, test_size=0.2, random_state=0)
        x = x_
        y = y_

        cls = LogisticRegression(max_iter=10000)
        cls.fit(x_train, y_train)

        #         Массив предсказанных значений
        y_pred = cls.predict(x_test)

        print('Accuracy = ', metrics.accuracy_score(y_test, y_pred))
        print('Precision = ', metrics.precision_score(y_test, y_pred, average='macro'))
        print('Recall = ', metrics.recall_score(y_test, y_pred, average='macro'))
        print('Коэффициенты регрессии: ', cls.coef_)

        fig, ax = plt.subplots()
        dicc = {}

        print('\n Массив предсказанных значений:\n')
        for i in range(len(x_test)):
            dicc[f'{x_test[i]}'] = y_pred[i]
        print(dicc)

        plt.ylabel('Feature №1')
        plt.xlabel('Feature №0')

        z = [(-0.7, -0.3, 0.33)]
        xfit = np.linspace(-4, 2.5)
        plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.viridis, edgecolors="k")
        for m, b, d in z:
            yfit = m * xfit + b

        plt.plot(xfit, yfit, '-k')
        stop = time.time()

        # metrics.confusion_matrix(y_test, y_pred)
        # class_names = [0, 1]
        # fig, ax = plt.subplots()
        # ticks = np.arange(len(class_names))
        # plt.xticks(ticks, class_names)
        # plt.yticks(ticks, class_names)
        # sns.heatmap(pd.DataFrame(metrics.confusion_matrix(y_test, y_pred)), annot=True)
        # plt.ylabel('Действительные значения')
        # plt.xlabel('Предсказанные значения')

        print('Время работы алгоритма: ', stop - start)
        plt.show()

    def logistic_rbf(x_, y_):

        """
        Функция, реализующая модель классификации на два класса на основе логистической регрессии с радиальными базисными функциями.

        Parameters
        ----------
        x_, y_: numpy array
            Массивы обучающей переменной и предсказываемой
        Returns
        -----------
        Accuracy, Precision, Recall, Коэффициенты регрессии, Массив предсказанных значений, График классификации
        """
        start = time.time()

        #         Массивы обучающей переменной и предсказываемой, разделенные на обучающую и тестовую выборки
        x_train, x_test, y_train, y_test = train_test_split(x_, y_, test_size=0.2, random_state=0)
        x = x_
        y = y_

        cls = SVC(kernel='rbf')
        cls.fit(x_train, y_train)
        #         Массив предсказанных значений
        y_pred = cls.predict(x_test)

        print('Accuracy = ', metrics.accuracy_score(y_test, y_pred))
        print('Precision = ', metrics.precision_score(y_test, y_pred, average='macro'))
        print('Recall = ', metrics.recall_score(y_test, y_pred, average='macro'))

        fig, ax = plt.subplots()
        dicc = {}

        print('\n Массив предсказанных значений:\n')
        for i in range(len(x_test)):
            dicc[f'{x_test[i]}'] = y_pred[i]
        print(dicc)

        plt.ylabel('Feature №1')
        plt.xlabel('Feature №0')

        z = [(-0.7, -0.3, 0.33)]
        xfit = np.linspace(-4, 2.5)
        plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.viridis, edgecolors="k")
        for m, b, d in z:
            yfit = m * xfit + b

        plt.plot(xfit, yfit, '-k')
        stop = time.time()

        # metrics.confusion_matrix(y_test, y_pred)
        # class_names = [0, 1]
        # fig, ax = plt.subplots()
        # ticks = np.arange(len(class_names))
        # plt.xticks(ticks, class_names)
        # plt.yticks(ticks, class_names)
        # sns.heatmap(pd.DataFrame(metrics.confusion_matrix(y_test, y_pred)), annot=True)
        # plt.ylabel('Действительные значения')
        # plt.xlabel('Предсказанные значения')

        print('Время работы алгоритма: ', stop - start)
        plt.show()

    def lin_with_l1(x_, y_):

        """
        Функция, реализующая модель классификации на два класса на основе логистической регрессии с регуляризацией L1
        Parameters
        ----------
        x_, y_: numpy array
            Массивы обучающей переменной и предсказываемой
        Returns
        -----------
        Accuracy, Precision, Recall, Коэффициенты регрессии, Массив предсказанных значений, График классификации
        """
        start = time.time()
        #         Массивы обучающей переменной и предсказываемой, разделенные на обучающую и тестовую выборки
        x_train, x_test, y_train, y_test = train_test_split(x_, y_, test_size=0.2, random_state=0)
        x = x_
        y = y_

        cls = LogisticRegression(penalty='l1', solver='liblinear')
        cls.fit(x_train, y_train)

        #         Массив предсказанных значений
        y_pred = cls.predict(x_test)

        print('Accuracy = ', metrics.accuracy_score(y_test, y_pred))
        print('Precision = ', metrics.precision_score(y_test, y_pred, average='macro'))
        print('Recall = ', metrics.recall_score(y_test, y_pred, average='macro'))

        fig, ax = plt.subplots()
        dicc = {}
        print('Коэффииценты регрессии: ', cls.coef_)

        print('\n Массив предсказанных значений:\n')
        for i in range(len(x_test)):
            dicc[f'{x_test[i]}'] = y_pred[i]
        print(dicc)

        plt.ylabel('Feature №1')
        plt.xlabel('Feature №0')

        z = [(-0.7, -0.3, 0.33)]
        xfit = np.linspace(-4, 2.5)
        plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.viridis, edgecolors="k")
        for m, b, d in z:
            yfit = m * xfit + b

        plt.plot(xfit, yfit, '-k')
        stop = time.time()

        # metrics.confusion_matrix(y_test, y_pred)
        # class_names = [0, 1]
        # fig, ax = plt.subplots()
        # ticks = np.arange(len(class_names))
        # plt.xticks(ticks, class_names)
        # plt.yticks(ticks, class_names)
        # sns.heatmap(pd.DataFrame(metrics.confusion_matrix(y_test, y_pred)), annot=True)
        # plt.ylabel('Действительные значения')
        # plt.xlabel('Предсказанные значения')

        print('Время работы алгоритма: ', stop - start)
        plt.show()

    def opornieVectora(dictionary):

        """
        Функция, реализующая модель классификации на два класса на основе метода опорных векторов.
        Parameters
        ----------
        dictionary: dictionary
            Словарь со всеми переменными, которые ранее были введены в функции vvod
        Returns
        -----------
        dicc: dict
            Словарь вида {x_test: y_pred}, тестовое значение переменной х сопоставляется с предсказанным значением переменной y
        """
        start = time.time()

        X_train = dictionary['X_train']
        X_test = dictionary['X_test']
        y_train = dictionary['y_train']
        y_test = dictionary['y_test']
        X = dictionary['X']
        y = dictionary['y']

        clf = SVC().fit(X_train, y_train)
        y_pred = clf.predict(X_test)

        print('Accuracy = ', metrics.accuracy_score(y_test, y_pred))
        print('Precision = ', metrics.precision_score(y_test, y_pred, average='macro'))
        print('Recall = ', metrics.recall_score(y_test, y_pred, average='macro'))

        fig, ax = plt.subplots()
        dicc = {}
        print('Коэффииценты регрессии: ', clf.dual_coef_)
        print('\n Массив предсказанных значений:\n')
        for i in range(len(X_test)):
            dicc[f'{X_test[i]}'] = y_pred[i]
        print(dicc)

        plt.ylabel('Feature №1')
        plt.xlabel('Feature №0')

        z = [(-0.7, -0.3, 0.33)]
        xfit = np.linspace(-4, 2.5)
        plt.scatter(X[:, 0], X[:, 1], c=y, cmap=plt.cm.viridis, edgecolors="k")
        for m, b, d in z:
            yfit = m * xfit + b

        plt.plot(xfit, yfit, '-k')

        stop = time.time()

        print('Время работы алгоритма: ', stop - start)
        plt.show()

    def funcForInput():

        """
        Функция для ввода данных, с помощью которых будет проводиться классификация
        Parameters
        ----------
        X: numpy.ndarray
            Массив значений переменной x
        y: numpy.ndarray
            Массив значений переменной y
        X_train: numpy.ndarray
            Массив обучающей выборки переменной x
        X_test: numpy.ndarray
            Массив тестовой выборки переменной x
        y_train: numpy.ndarray
            Массив обучающей выборки переменной y
        y_test: numpy.ndarray
            Массив тестовой выборки переменной y
        Returns
        -----------
        Final: dictionary
            Словарь со всеми переменными, которые ранее были введены, нужен для ввода в основную функцию
        """

        from sklearn.datasets import make_classification
        from sklearn.model_selection import train_test_split
        X, y = make_classification(n_samples=100, n_features=4, class_sep=0.98, random_state=0)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
        Final = {'X_train': X_train,
                 'X_test': X_test,
                 'y_train': y_train,
                 'y_test': y_test,
                 'X': X,
                 'y': y, }
        return Final


import numpy as np
from typing import Callable, Optional
import autograd.numpy as npa
from autograd import grad, hessian
import numexpr as ne
import sympy as sp
from matplotlib import pyplot as plt
import copy


from .input_validation import check_expression, check_restr, check_point
from .prepocessing import prepare_all


class LogBarrirers:
    """
       Метод решения задачи методом логарифмических барьеров.

       Parameters
       ----------
       function: Callable
           Функция для минимизации.

       restr_uneq: list
           Список вызываемых функций питона, которые имеют смысл левой части ограничений g(x)  0


       x0: np.ndarray
           Плоский массив, содержащий начальную точку.

       eps: Optional[float] = 1e-12
           Точность. От этого параметра зависит, когда будет остановлен итерационные процесс поиска решения. При
           слишком малых значениях, есть вероятность бесконечного цикла или nan в ответе.
       """

    def __init__(self,
                 function: Callable,
                 restrictions: list,
                 x0: np.ndarray,
                 eps: Optional[float] = 1e-12):
        self.function = function
        self.restrictions = restrictions
        self.x = x0
        self.eps = eps
        self.mu = 1

    def solve(self):
        l_hess = hessian(self.lagrange)
        l_grad = grad(self.lagrange)
        step = np.linalg.inv(l_hess(self.x)) @ l_grad(self.x)

        while not np.allclose(step, np.zeros(self.x.shape)):
            self.x = self.x - step
            step = np.linalg.inv(l_hess(self.x)) @ l_grad(self.x)
            self.mu *= 10
        return self.x

    def lagrange(self, x):
        l = self.function(x)*self.mu
        for i in self.restrictions:
            l -= npa.log(i(x))
        return l


if __name__ == '__main__':
    f = 'x1**2 + x2**2 + (0.5*1*x1 + 0.5*2*x2)**2 + (0.5*1*x1 + 0.5*2*x2)**4'
    f_graph = copy.deepcopy(f)
    subject_to = 'x1+x2<=0;2*x1-3*x2<=1'
    # point_min = np.array([0, 0])
    # point_start = np.array([-5, 4.])
    point_start = '-5;4'

    # input_validation
    f = check_expression(f)
    subject_to = check_restr(subject_to, method='log_barrier')
    point_start = check_point(point_start, f, subject_to, 'log_barrier')
    # preprocessing
    f, subject_to, point_start = prepare_all(f, subject_to, 'log_barrier', point_start)
    # solver
    task = LogBarrirers(f, subject_to, point_start)
    ans = task.solve()
    # print(np.allclose(ans, point_min))
    print(ans)

    # Рисуем график
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_title('Метод логарифмических барьеров ')
    ax.set_zlabel('F(x1,x2)')
    ax.set_ylabel('x2')
    ax.set_xlabel('x1')

    x1 = np.linspace(-1, 1, 25)
    x2 = np.linspace(-1, 1, 25)
    x1, x2 = np.meshgrid(x1, x2)
    Z = ne.evaluate(f_graph)

    surf = ax.plot_surface(x1, x2, Z, cmap='viridis', shade=True, alpha=0.5)
    fig.colorbar(surf, shrink=0.5, aspect=10)

    func1 = sp.sympify(f_graph)

    x1, x2 = sp.symbols('x1 x2')

    ax.scatter(float(ans[0]), float(ans[1]),
               func1.subs([(x1, float(ans[0])), (x2, float(ans[1]))]),
               c='r', s=50,
               label=f'Точка экстремума \n x1 = {float(ans[0])},\n x2 = {float(ans[1])},\n'
                     f' F(x1,x2) = {func1.subs([(x1, float(ans[0])), (x2, float(ans[1]))])}')

    ax.legend(labelcolor='black')
    plt.show()



import numpy as np
from copy import deepcopy
from itertools import product
import time
import plotly.graph_objects as go
from matplotlib import pyplot as plt

MAX_MODE = 'MAX'  # режим максимизации
MIN_MODE = 'MIN'  # режим минимизации


class SimplexMethod:
    """
    Класс, реализующий симплекс-метод и дальнейшее решение по методу Гомори.

    """

    def __init__(self, c, a, b, mode):
        """
        Функция для инициализации всех переменных.

        """

        self.main_variables_count = a.shape[1]  # количество переменных
        self.restrictions_count = a.shape[0]  # количество ограничений
        self.variables_count = self.main_variables_count + self.restrictions_count  # количество переменных
        self.mode = mode  # запоминаем режим работы

        self.c = np.concatenate([c, np.zeros((self.restrictions_count + 1))])  # коэффициенты функции
        self.f = np.zeros((self.variables_count + 1))  # значения функции F
        self.basis = [i + self.main_variables_count for i in
                      range(self.restrictions_count)]  # индексы базисных переменных
        self.task = {'a': deepcopy(a), 'b': deepcopy(b), 'c': deepcopy(c)}  # сохраняем начальную задачу

        self.init_table(a, b)

    # инициализация таблицы
    def init_table(self, a, b):
        """
        Функция для инициализации таблицы.

        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        self.table = np.zeros((self.restrictions_count, self.variables_count + 1))  # коэффициенты таблицы

        for i in range(self.restrictions_count):
            for j in range(self.main_variables_count):
                self.table[i][j] = a[i][j]

            for j in range(self.restrictions_count):
                self.table[i][j + self.main_variables_count] = int(i == j)

            self.table[i][-1] = b[i]

    # получение строки с максимальным по модулю отрицательным значением b
    def get_negative_b_row(self):
        """
        Функция получения строки с максимальным по модулю отрица-тельным значением b.

        Parameters
        -----------
        row: numpy.array
        Строка с коэффициентами.

        Returns
        -----------
        row: numpy.array
        Строка с максимальным по модулю отрицательным значением b.

        """
        row = -1

        for i, a_row in enumerate(self.table):
            if a_row[-1] < 0 and (row == -1 or abs(a_row[-1]) > abs(self.table[row][-1])):
                row = i

        return row

    # получение столбца с максимальным по модулю элементом в строке
    def get_negative_b_column(self, row):

        """
        Функция получения столбца с максимальным по модулю эле-ментом в строке.

        Parameters
        -----------
        column: numpy.array
        Столбец с коэффициентами.

        Returns
        -----------
        column: numpy.array
        Столбец с максимальным по модулю элементом в столбце.

        """
        column = -1

        for i, aij in enumerate(self.table[row][:-1]):
            if aij < 0 and (column == -1 or abs(aij) > abs(self.table[row][column])):
                column = i

        return column

    # удаление отрицательных свободных коэффициентов
    def remove_negative_b(self):
        """
        Функция удаления отрицательных свободных коэффициентов.

        Parameters
        -----------
        column: numpy.array
        Столбец с коэффициентами.

        row: numpy.array
        Строка с коэффициентами.

        """
        while True:
            row = self.get_negative_b_row()  # ищем строку, в которой находится отрицательное b

            if row == -1:  # если не нашли такую строку
                return True  # то всё хорошо

            column = self.get_negative_b_column(row)  # ищем разрешающий столбец

            if column == -1:
                return False  # не удалось удалить

            self.gauss(row, column)  # выполняем исключение гаусса
            self.calculate_f()

    # выполнение шага метода гаусса
    def gauss(self, row, column):
        """
        Функция выполняющая шаг метода Гаусса.

        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        self.table[row] /= self.table[row][column]

        for i in range(self.restrictions_count):
            if i != row:
                self.table[i] -= self.table[row] * self.table[i][column]

                for j in range(self.variables_count):
                    if abs(self.table[i][j]) < 1e-10:
                        self.table[i][j] = 0

        self.basis[row] = column  # делаем переменную базисной

    # расчёт значений F
    def calculate_f(self):
        """
        Функция, считающая значения заданной функции.

        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        for i in range(self.variables_count + 1):
            self.f[i] = -self.c[i]

            for j in range(self.restrictions_count):
                self.f[i] += self.c[self.basis[j]] * self.table[j][i]

    # расчёт симплекс-отношений для столбца column
    def get_relations(self, column):
        """
        Функция, делающая расчет симплекс отношений для столбца column.

        Parameters
        -----------
        q: list
        Список симплекс-отношений.

        Returns
        -----------
        q: list
        Список со значениями симплекс-отношений.

        """

        q = []

        for i in range(self.restrictions_count):
            if self.table[i][column] == 0:
                q.append(np.inf)
            else:
                q_i = self.table[i][-1] / self.table[i][column]
                q.append(q_i if q_i >= 0 else np.inf)

        return q

    # получение решения
    def get_solve(self):
        x = np.zeros((self.variables_count))

        # заполняем решение
        for i in range(self.restrictions_count):
            x[self.basis[i]] = self.table[i][-1]

        return x  # возвращаем полученное решение

    # получение первого вещественного решения
    def get_first_real(self, x):
        return next((i for i, xi in enumerate(x[:self.main_variables_count]) if xi != int(xi)), -1)

    # решение
    def solve(self, debug):
        """
        Функция, решающая симплекс-таблицу и выводящая последова-тельность итераций.

        Parameters
        -----------
        iteration: int
        Номер итерации.

        Returns
        -----------
        True, False: bool

        """
        self.calculate_f()
        s = ''
        if debug:
            s += '\nIteration 0'
            s += self.print_table()

        if not self.remove_negative_b():
            s += 'Решения нет'
            return False

        iteration = 1

        while True:
            self.calculate_f()

            if debug:
                s += f'\nIteration,{iteration}'
                s += self.print_table()

            if all(fi >= 0 if self.mode == MAX_MODE else fi <= 0 for fi in self.f[:-1]):  # если план оптимален
                break  # то завершаем работу

            column = (np.argmin if self.mode == MAX_MODE else np.argmax)(self.f[:-1])  # получаем разрешающий столбец
            q = self.get_relations(column)  # получаем симплекс-отношения для найденного столбца

            if all(qi == np.inf for qi in q):  # если не удалось найти разрешающую строку
                s += 'Решения нет'  # сообщаем, что решения нет
                return False

            self.gauss(np.argmin(q), column)  # выполняем исключение гаусса
            iteration += 1

        return True, s  # решение есть

    # получение дробной части
    def part(self, a):
        """
        Функция, возвращающая дробную часть числа.

        Returns
        -----------
        a: int
        Дробная часть числа.

        """
        if a > 0:
            return a - int(a)

        return int(abs(a)) + 1 + a

    def make_gamori(self, x, real_index, debug):
        """
        Функция, реализующая метод Гомори.

        Parameters
        -----------
        r_main: list
        r_basis: list
        task1_a: np.array()
        task1_b: np.array

        Returns
        -----------
        Целочисленные решения.

        """
        b = self.part(x[real_index])
        r_main = [0 if i in self.basis else self.part(self.table[real_index][i]) for i in
                  range(self.main_variables_count)]
        r_basis = [0 if i in self.basis else self.part(self.table[real_index][i]) for i in
                   range(self.main_variables_count, self.variables_count)]

        # добавляем условия
        task1_a = np.append(self.task['a'], np.array([r_main]), 0)
        task1_b = np.append(self.task['b'], -b)

        simplex1 = SimplexMethod(self.task['c'], task1_a, task1_b, self.mode)

        for i in range(self.restrictions_count):
            for j in range(self.variables_count):
                simplex1.table[i][j] = self.table[i][j]

            simplex1.table[i][-1] = self.table[i][-1]

        for i in range(self.main_variables_count):
            simplex1.table[-1][i] = -r_main[i]

        for i in range(self.main_variables_count, self.variables_count):
            simplex1.table[-1][i] = -r_basis[i - self.main_variables_count]

        return simplex1.solve_integer(debug)

    # получение целочисленных решений
    def solve_integer(self, debug=False):
        """
        Функция, получающая целочисленные решения.

        """
        s = ''
        print(self.print_task())

        # если решение не было найдено, то добавляем пустое решение
        if not self.solve(debug)[0]:
            return []

        s += f'Решение:, {self.get_solve()[:self.main_variables_count]}, , F:, {self.f[-1]}, \n'
        x = self.get_solve()
        real_index = self.get_first_real(x)

        # если решение содержало только целые числа, то возвращаем решение
        if real_index == -1:
            return [(x, self.f[-1])]
        print(s)
        return self.make_gamori(x, real_index, debug)

    # вывод лучшего решения
    def print_best_solve(self, solves):
        sign = 1 if self.mode == MAX_MODE else -1
        imax = 0

        for i, solve in enumerate(solves):
            if solve[1] * sign > solves[imax][1] * sign:
                imax = i

        x, f = solves[imax]
        print('\nОтвет: ', x[:self.main_variables_count], 'F:', f)
        return x[:self.main_variables_count], f

    # вывод симплекс-таблицы
    def print_table(self):
        """
        Функция, выводящая симплекс-таблицы на каждой итерации.

        """
        s = ''
        s += '     |' + ''.join(['   x%-3d |' % (i + 1) for i in range(self.variables_count)]) + '    b   |'

        for i in range(self.restrictions_count):
            s += '%4s |' % ('x' + str(self.basis[i] + 1)) + ''.join(
                [' %6.2f |' % aij for j, aij in enumerate(self.table[i])])

        s += '   F |' + ''.join([' %6.2f |' % aij for aij in self.f])
        # print('   x |' + ''.join([' %6.2f |' % xi for xi in self.get_solve()]))
        return s

    # вывод коэффициента
    def print_coef(self, ai, i):
        if ai == 1:
            return 'x%d' % (i + 1)

        if ai == -1:
            return '-x%d' % (i + 1)

        return '%.2fx%d' % (ai, i + 1)

    # вывод задачи
    def print_task(self, full=True):
        s = ' + '.join(['%.2fx%d' % (ci, i + 1) for i, ci in enumerate(self.c[:self.main_variables_count]) if
                        ci != 0]), '-> ', self.mode

        for row in self.table:
            if full:
                s += ' + '.join(
                    [self.print_coef(ai, i) for i, ai in enumerate(row[:self.variables_count]) if ai != 0]), '=', row[
                         -1]
            else:
                s += ' + '.join(
                    [self.print_coef(ai, i) for i, ai in enumerate(row[:self.main_variables_count]) if ai != 0]), '<=', \
                     row[-1]
        return s


def main(c, a, b, mode="MAX"):
    """
    Функции для считывания введенной функции и ограничений и решающая задачу методом Гомори.
    Parameters
    -----------
    с: numpy.array
    Коэффициенты заданной функции.
    a: numpy.array
    Коэффициенты заданных ограничений
    b: numpy.array
    Коэффициенты свободных членов ограничений
    mode: str
    {"MIN", "MAX"}
    Returns
    -----------
    Найденную координату и значение функции в этой координате.

    """

    simplex = SimplexMethod(c, a, b, mode)

    print("Решение методом Гомори:")
    solves = simplex.solve_integer(True)
    return simplex.print_best_solve(solves)


def Gomory_visualize(c, a, b, result):
    fig, ax = plt.subplots()
    plt.scatter(result[0][0], result[0][1])
    x2min = 0
    x2max = -np.inf
    for i in range(len(a)):
        g = lambda x1: (b[i] - a[i][0] * x1) / a[i][1]
        x = np.linspace(0, result[0][0] + 2)
        x2 = g(x)
        if np.min(x2) < x2min:
            x2min = np.min(x2)
        if np.max(x2) > x2max:
            x2max = np.max(x2)
        plt.plot(x, x2)
    plt.show()

# Тесты

# c = np.array([4, 5, 6])
#
# a = np.array([
#     [1, 2, 3],
#     [4, 3, 2],
#     [3, 1, 1]
# ])
# b = np.array([35, 45, 40])
#
# t1 = time.time()
# visualize(c, a, b, main(c, a, b, mode="MAX"))
# t2 = time.time()
# print('Время в секундах: ', t2-t1)


from sympy import *
from sklearn.preprocessing import PolynomialFeatures
import numpy as np
import matplotlib.pyplot as plt
import numexpr as ne
from .AdjGrad import *
from .brent import *


class PolRegres:
    def find(y, x, n, m, deg, reg, L, sigma):
        
        """
        Реализует модель полиномиальной регрессии 3 видами регуляторов: L1, L2 и Стьюдента.
        
        Parameters
        ===========
        :param y: np.array
            массив предсказываемых данных 
        :param x: list
            массив предикатов X
        :param n: int
            кол-во строк в матрице
        :param m: int
            кол-во столбцов в матрице
        :param deg: int
            степень полинома
        :param reg: str
            параметр, отвечающий за вид регуляризации
        :param L: int
            коэффициент регуляции
        :param sigma: int
            стандартное отклонение остатков
        :return:
            Возвращает вектор модельных предсказанных данных, массив коэффициентов регрессии, 
            свободный член, функция в аналитическом виде.
        """

        x = np.array(x)
        x = PolynomialFeatures(deg).fit_transform(x)

        m = len(x[0]) - 1

        X = 'x1'
        for i in range(m):
            X = X + " " + f'x{i + 2}'



        if reg == 'None':
            omega = np.array([list(symbols(X))])
            loss_function = (1 / (2 * n)) * sum((y - np.dot(x, omega.transpose())) ** 2)

        elif reg == 'L1':
            omega = np.array([list(symbols(X))])
            loss_function = (1 / (2 * n)) * sum((y - np.dot(x, omega.transpose())) ** 2) + L * (omega ** 2 / 2).sum()

        elif reg == 'L2':
            omega = np.array([list(symbols(X))])
            loss_function = (1 / (2 * n)) * sum((y - np.dot(x, omega.transpose())) ** 2) + L * sum(omega ** 2)

        elif reg == 'norm':
            omega = np.array([list(symbols(X))])
            loss_function = (1 / (2 * n)) * sum((y - np.dot(x, omega.transpose())) ** 2) + (1 / (2 * sigma)) * sum(
                omega ** 2)



        omega_point = [0 for i in range(m + 1)]
        omega_min = AdjGrad.find(str(loss_function[0]), omega_point, 0.00001, 0, 0, 0)

        y_ = np.dot(x, omega_min)

        print(f'Вектор модельных предсказанных данных: {y_}') 
        print(f'Массив коэффициентов регрессии: {omega_min[1:]}') 
        print(f'Свободный член: {omega_min[0]}') 
        y_str = str(omega_min[0]) 
        for i in range(1, len(omega_min)): 
            y_str = y_str + f' + {omega_min[i]} * {list(symbols(X))[i-1]}' 
        print(f'Функция в аналитическом виде: y^={y_str}')
        
        if (len(omega_min) - 1) == 1:
            fig, ax = plt.subplots(figsize=(10, 10))

            ax.set_title('Экспоненциальная регрессия', loc='left', fontweight='bold')
            ax.set_title(f'y^ = {y_str}', loc='right')
            ax.set_ylabel('y^')
            ax.set_xlabel('x1')

            x1 = np.linspace(-10, 10, 50)
            y = ne.evaluate(y_str)

            plt.plot(x1, y, color='b')
            x1 = Symbol('x1')
            for i in range(len(x)):
                ax.scatter(x[i][1], sympify(y_str).subs(x1, x[i][1]), c='r')

            plt.show()

        elif (len(omega_min) - 1) == 2:
            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(projection='3d')
            ax.set_title(f'y^ = {y_str}', loc='right')
            ax.set_zlabel('y^')
            ax.set_ylabel('x2')
            ax.set_xlabel('x1')

            x1 = np.linspace(np.min(x), np.max(x), 25)
            x2 = np.linspace(np.min(x), np.max(x), 25)
            x1, x2 = np.meshgrid(x1, x2)
            y = ne.evaluate(y_str)

            surf = ax.plot_surface(x1, x2, y, cmap='viridis', shade=True, alpha=0.5)
            fig.colorbar(surf, shrink=0.5, aspect=10)
            x1, x2 = symbols('x1 x2')
            for i in range(len(x)):
                ax.scatter(x[i][1], x[i][2], sympify(y_str).subs([(x1, x[i][1]), (x2, x[i][2])]), c='r')

            plt.show()

# y = []
# print("Введите кол-во строк в матрице(n). Например 3")
# n = int(input())
# print("Введите кол-во столбцов в матрице(n). Например 3")
# m = int(input())
# print("Введите массив предсказываемых данных y = (y1, y2, ..., yn)")
# for i in range(n):
#    print(f'Введите {i} значение. Например 1')
#    y.append(float(input()))
# print("Введите массив предикантов X размерностью n x m")
# X = []
# for i in range(n):
#    xij = []
#    for j in range(m):
#        print(f'Введите x{i, j}')
#        xij.append(float(input()))
#    X.append(xij)
#
# L = 0.95
# sigma = 0.1
#
# print("Введите степень полинома (целое число больше нуля). Например: 2")
# deg = int(input())
# q = -1
#
# print("Хотите ввести параметр, отвечающий за вид регуляризации? 1 - да / 0 - нет")
# q = -1
# while q < 0 or q > 1:
#     q = int(input())
#     if q < 0 or q > 1:
#         print("Неправильный ввод. 1 - да / 0 - нет")
#
# if q == 1:
#     print("Введите L1, или L2, или norm ")
#     reg = 'j'
#     while reg != "L1" and reg != "L2" and reg != "norm":
#         reg = input()
#         if reg != "L1" and reg != "L2" and reg != "norm":
#             print("Такой комманды нет. Введите снова. L1, или L2, или norm")
#
#     if reg == "L1" or reg == "L2":
#         print("Введите коэффициент регуляции. Например 0.95. Значение должно быть >=0. (чем больше, тем сильнее регуляризация)")
#         L = -1
#         while L < 0 or L > 1:
#             L = float(input())
#             if L < 0 or L > 1:
#                 print("Некорректно введен коэффициент регуляции. Значение должно быть >=0.")
#
#     elif reg == "norm":
#         print("Введите предполагаемое стандартное отклонение остатков. Например 0.1. Значение должно быть >=0 и <=1. (чем больше, тем слабее регуляризация)")
#         sigma = -1
#         while sigma < 0 or sigma > 1:
#             sigma = float(input())
#             if sigma < 0 or sigma > 1:
#                 print("Некорректно введено предполагаемое стандартное отклонение остатков. Значение должно быть >=0 и <=1.")
# elif q == 0:
#     reg = 'None'
#
# functionss = PolRegres()
# functionss.find(np.array(y).reshape(-1, 1), X, n, m, deg, reg, L, sigma)

#functionss = PolRegres()
#functionss.find(np.array([1, 1, 1]).reshape(-1, 1), [[0], [-1], [-1]], 3, 1, 2, 'None', 0.95, 0.1)

#functionss = PolRegres()
#functionss.find(np.array([1, 1, 1]).reshape(-1, 1), [[0], [-1], [-1]], 3, 1, 1, 'None', 0.95, 0.1)



import time
import numpy as np
from copy import deepcopy
from itertools import product

from matplotlib import pyplot as plt

MAX_MODE = 'MAX'
MIN_MODE = 'MIN'


class SimplexMethod:
    """
    Класс, реализующий симплекс-метод и дальнейшее решение по методу ветвей и границ.

    """

    def __init__(self, c, a, b, mode):
        """
    Функция для инициализации всех переменных.

        """
        self.main_variables_count = a.shape[1]  # количество переменных
        self.restrictions_count = a.shape[0]  # количество ограничений
        self.variables_count = self.main_variables_count + self.restrictions_count  # количество переменных
        self.mode = mode  # запоминаем режим работы

        self.c = np.concatenate([c, np.zeros((self.restrictions_count + 1))])  # коэффициенты функции
        self.f = np.zeros((self.variables_count + 1))  # значения функции F
        self.basis = [i + self.main_variables_count for i in
                      range(self.restrictions_count)]  # индексы базисных переменных
        self.indent = 0
        self.task = {'a': deepcopy(a), 'b': deepcopy(b), 'c': deepcopy(c)}  # сохраняем начальную задачу

        self.init_table(a, b)

    # инициализация таблицы
    def init_table(self, a, b):
        """
        Функция для инициализации таблицы.
        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        self.table = np.zeros((self.restrictions_count, self.variables_count + 1))  # коэффициенты таблицы

        for i in range(self.restrictions_count):
            for j in range(self.main_variables_count):
                self.table[i][j] = a[i][j]
                self.table[i][j + self.main_variables_count] = int(i == j)

            self.table[i][-1] = b[i]

    # получение строки с максимальным по модулю отрицательным значением b
    def get_negative_b_row(self):
        """
        Функция получения строки с максимальным по модулю отрица-тельным значением b.
        Parameters
        -----------
        row: numpy.array
        Строка с коэффициентами.

        Returns
        -----------

        row: numpy.array
        Строка с максимальным по модулю отрицательным значением b.

        """
        row = -1

        for i, a_row in enumerate(self.table):
            if a_row[-1] < 0 and (row == -1 or abs(a_row[-1]) > abs(self.table[row][-1])):
                row = i

        return row

    # получение столбца с максимальным по модулю элементом в строке
    def get_negative_b_column(self, row):
        """
        Функция получения столбца с максимальным по модулю эле-ментом в строке.
        Parameters
        -----------
        column: numpy.array
        Столбец с коэффициентами.

        Returns
        -----------
        column: numpy.array
        Столбец с максимальным по модулю элементом в столбце.

        """
        column = -1

        for i, aij in enumerate(self.table[row][:-1]):
            if aij < 0 and (column == -1 or abs(aij) > abs(self.table[row][column])):
                column = i

        return column

    # удаление отрицательных свободных коэффициентов
    def remove_negative_b(self):
        """
        Функция удаления отрицательных свободных коэффициентов.
        Parameters
        -----------
        column: numpy.array
        Столбец с коэффициентами.
        row: numpy.array
        Строка с коэффициентами.

        """
        while True:
            row = self.get_negative_b_row()  # ищем строку, в которой находятся отрицательное b

            if row == -1:  # если не нашли такую строку
                return True  # то всё хорошо

            column = self.get_negative_b_column(row)  # ищем разрешающий столбец

            if column == -1:
                return False  # не удалось удалить

            self.gauss(row, column)  # выполняем исключение гаусса
            self.calculate_f()
            self.print_table()

    # выполнение шага метода гаусса
    def gauss(self, row, column):
        """
        Функция выполняющая шаг метода Гаусса.
        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        self.table[row] /= self.table[row][column]

        for i in range(self.restrictions_count):
            if i != row:
                self.table[i] -= self.table[row] * self.table[i][column]

        self.basis[row] = column  # делаем переменную базисной

    # расчёт значений F
    def calculate_f(self):
        """
        Функция, считающая значения заданной функции.
        Parameters
        -----------
        table: numpy.array
        Таблица с коэффициентами.

        """
        for i in range(self.variables_count + 1):
            self.f[i] = -self.c[i]

            for j in range(self.restrictions_count):
                self.f[i] += self.c[self.basis[j]] * self.table[j][i]

    # расчёт симплекс-отношений для столбца column
    def get_relations(self, column):
        """
        Функция, делающая расчет симплекс отношений для столбца column.
        Parameters
        -----------
        q: list
        Список симплекс-отношений.

        Returns
        -----------
        q: list
        Список со значениями симплекс-отношений.

        """
        q = []

        for i in range(self.restrictions_count):
            if self.table[i][column] == 0:
                q.append(np.inf)
            else:
                q_i = self.table[i][-1] / self.table[i][column]
                q.append(q_i if q_i >= 0 else np.inf)

        return q

    # получение решения
    def get_solve(self):
        x = np.zeros((self.variables_count))

        # заполняем решение
        for i in range(self.restrictions_count):
            x[self.basis[i]] = self.table[i][-1]

        return x  # возвращаем полученное решение

    # получение первого вещественного решения
    def get_first_real(self, x):
        return next((i for i, xi in enumerate(x[:self.main_variables_count]) if xi != int(xi)), -1)

    # решение
    def solve(self, debug):
        """
        Функция, решающая симплекс-таблицу и выводящая последовательность итераций.
        Parameters
        -----------
        iteration: int
        Номер итерации.

        Returns
        -----------
        True, False: bool

        """
        self.calculate_f()

        if debug:
            print('\nИтерация 0')
            self.print_table()

        if not self.remove_negative_b():
            print(self.indent, 'Решения нет')
            return False

        iteration = 1

        while True:
            self.calculate_f()

            if debug:
                print('\nИтерация', iteration)
                self.print_table()

            if all(fi >= 0 if self.mode == MAX_MODE else fi <= 0 for fi in self.f[:-1]):  # если план оптимален
                break  # то завершаем работу

            column = (np.argmin if self.mode == MAX_MODE else np.argmax)(self.f[:-1])  # получаем разрешающий столбец
            q = self.get_relations(column)  # получаем симплекс-отношения для найденного столбца

            if all(qi == np.inf for qi in q):  # если не удалось найти разрешающую строку
                print('Решения нет')  # сообщаем, что решения нет
                return False

            self.gauss(np.argmin(q), column)  # выполняем исключение гаусса
            iteration += 1

        return True  # решение есть

    # получение целочисленных решений
    def solve_integer(self, debug=False, depth=0):
        """
        Функция, получающая целочисленные решения путем разбиения задачи на две ветки.

        """

        self.indent = '    ' * depth
        print(self.indent + 'Решение:')
        self.print_task()

        # если решение не было найдено, то добавляем пустое решение
        if not self.solve(debug):
            return []

        print(self.indent, "Решение:", self.get_solve()[:self.main_variables_count], ', F:', self.f[-1], '\n')
        x = self.get_solve()
        real_index = self.get_first_real(x)

        # если решение содержало только целые числа, то возвращаем решение
        if real_index == -1:
            return [(x, self.f[-1])]

        b = int(x[real_index])  # получаем новое условие

        # добавляем условия
        task1_a = np.append(self.task['a'],
                            np.array([[1 if i == real_index else 0 for i in range(self.main_variables_count)]]), 0)
        task1_b = np.append(self.task['b'], b)

        task2_a = np.append(self.task['a'],
                            np.array([[-1 if i == real_index else 0 for i in range(self.main_variables_count)]]), 0)
        task2_b = np.append(self.task['b'], -(b + 1))

        # разбиваем задачу на 2 ветки решения
        simplex1 = SimplexMethod(self.task['c'], task1_a, task1_b, self.mode)
        simplex2 = SimplexMethod(self.task['c'], task2_a, task2_b, self.mode)

        print(self.indent, 'Divide to tasks: x%d <= %.3f and x%d >= %.3f' % (real_index + 1, b, real_index + 1, b + 1))

        # запускаем решение для каждой ветки
        solve1 = simplex1.solve_integer(debug, depth + 1)
        solve2 = simplex2.solve_integer(debug, depth + 1)

        return solve1 + solve2

    # вывод лучшего решения
    def print_best_solve(self, solves):
        sign = 1 if self.mode == MAX_MODE else -1
        imax = 0

        for i, solve in enumerate(solves):
            if solve[1] * sign > solves[imax][1] * sign:
                imax = i

        x, f = solves[imax]
        print('\nЛучшее решение: ', x[:self.main_variables_count], 'F:', f)
        return x[:self.main_variables_count], f

    # проверка выполнения ограничений
    def check_solve(self, solve):
        """
        Функция, выполняющая проверку выполнения ограничений.

        Returns
        -----------
        True, False: bool

        """

        for row, bi in zip(self.task['a'], self.task['b']):
            if sum(ai * xi for ai, xi in zip(row, solve)) > bi:
                return False

        return True

    # перебор
    def solve_integer_bruteforce(self, x_max=20):
        solves = []
        print('Valid solutions (bruteforce):')

        # идём по всем наборам x
        for x in product(range(x_max), repeat=self.main_variables_count):
            if self.check_solve(x):  # если решение подходит
                f = sum(xi * ci for xi, ci in zip(x, self.task['c']))
                print(x[:self.main_variables_count], 'F:', f)
                solves.append((x, f))

        self.print_best_solve(solves)

    # вывод симплекс-таблицы
    def print_table(self):
        """
        Функция, выводящая симплекс-таблицы на каждой итерации.

        """
        print('     |' + ''.join(['   x%-3d |' % (i + 1) for i in range(self.variables_count)]) + '    b   |')

        for i in range(self.restrictions_count):
            print('%4s |' % ('x' + str(self.basis[i] + 1)) + ''.join(
                [' %6.2f |' % aij for j, aij in enumerate(self.table[i])]))

        print('   F |' + ''.join([' %6.2f |' % aij for aij in self.f]))
        print('   x |' + ''.join([' %6.2f |' % xi for xi in self.get_solve()]))

    # вывод коэффициента
    def print_coef(self, ai, i):
        if ai == 1:
            return 'x%d' % (i + 1)

        if ai == -1:
            return '-x%d' % (i + 1)

        return '%.2fx%d' % (ai, i + 1)

    # вывод задачи
    def print_task(self, full=False):
        print(self.indent, ' + '.join(
            ['%.2fx%d' % (ci, i + 1) for i, ci in enumerate(self.c[:self.main_variables_count]) if ci != 0]), '-> ',
              self.mode)

        for row in self.table:
            if full:
                print(self.indent, ' + '.join(
                    [self.print_coef(ai, i) for i, ai in enumerate(row[:self.variables_count]) if ai != 0]), '=',
                      row[-1])
            else:
                print(self.indent, ' + '.join(
                    [self.print_coef(ai, i) for i, ai in enumerate(row[:self.main_variables_count]) if ai != 0]), '<=',
                      row[-1])


def bb(c, a, b, mode):
    """
    Функции для считывания введенной функции и ограничений и решающая задачу методом ветвей и границ.

    Parameters
    -----------
    с: numpy.array
    Коэффициенты заданной функции.
    a: numpy.array
    Коэффициенты заданных ограничений
    b: numpy.array
    Коэффициенты свободных членов ограничений
    mode: str
    {"MIN", "MAX"}
    Returns
    -----------
    Найденную координату и значение функции в этой координате.
    """

    simplex = SimplexMethod(c, a, b, mode)

    simplex.solve_integer_bruteforce()

    print("\n\nSolve with branch and bound:")
    solves = simplex.solve_integer(True)
    return simplex.print_best_solve(solves)


def VG_visualize(c, a, b, result):
    fig, ax = plt.subplots()
    plt.scatter(result[0][0], result[0][1])
    x2min = 0
    x2max = -np.inf
    for i in range(len(a)):
        g = lambda x1: (b[i] - a[i][0] * x1) / a[i][1]
        x = np.linspace(0, result[0][0] + 2)
        x2 = g(x)
        if np.min(x2) < x2min:
            x2min = np.min(x2)
        if np.max(x2) > x2max:
            x2max = np.max(x2)
        plt.plot(x, x2)
    plt.show()


# Тесты

# c = np.array([4, 5, 6])
# a = np.array([
#     [1, 2, 3],
#     [4, 3, 2],
#     [3, 1, 1]
# ])
# b = np.array([35, 45, 40])
# t1 = time.time()
# visualize(c, a, b, bb(c, a, b, mode="MAX"))
# t2 = time.time()
# print('Время в секундах: ', t2 - t1)
